@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2010 Intel Corporation.
+ * Copyright © 2008-2012 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
@@ -111,24 +111,26 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	}
 
  retry:
-	openconnect_SSL_printf(vpninfo->https_ssl, "CONNECT /CSCOSSLC/tunnel HTTP/1.1\r\n");
-	openconnect_SSL_printf(vpninfo->https_ssl, "Host: %s\r\n", vpninfo->hostname);
-	openconnect_SSL_printf(vpninfo->https_ssl, "User-Agent: %s\r\n", vpninfo->useragent);
-	openconnect_SSL_printf(vpninfo->https_ssl, "Cookie: webvpn=%s\r\n", vpninfo->cookie);
-	openconnect_SSL_printf(vpninfo->https_ssl, "X-CSTP-Version: 1\r\n");
-	openconnect_SSL_printf(vpninfo->https_ssl, "X-CSTP-Hostname: %s\r\n", vpninfo->localname);
+	openconnect_SSL_printf(vpninfo, "CONNECT /CSCOSSLC/tunnel HTTP/1.1\r\n");
+	openconnect_SSL_printf(vpninfo, "Host: %s\r\n", vpninfo->hostname);
+	openconnect_SSL_printf(vpninfo, "User-Agent: %s\r\n", vpninfo->useragent);
+	openconnect_SSL_printf(vpninfo, "Cookie: webvpn=%s\r\n", vpninfo->cookie);
+	openconnect_SSL_printf(vpninfo, "X-CSTP-Version: 1\r\n");
+	openconnect_SSL_printf(vpninfo, "X-CSTP-Hostname: %s\r\n", vpninfo->localname);
 	if (vpninfo->deflate)
-		openconnect_SSL_printf(vpninfo->https_ssl, "X-CSTP-Accept-Encoding: deflate;q=1.0\r\n");
-	openconnect_SSL_printf(vpninfo->https_ssl, "X-CSTP-MTU: %d\r\n", vpninfo->mtu);
-	openconnect_SSL_printf(vpninfo->https_ssl, "X-CSTP-Address-Type: %s\r\n",
+		openconnect_SSL_printf(vpninfo, "X-CSTP-Accept-Encoding: deflate;q=1.0\r\n");
+	openconnect_SSL_printf(vpninfo, "X-CSTP-MTU: %d\r\n", vpninfo->mtu);
+	openconnect_SSL_printf(vpninfo, "X-CSTP-Address-Type: %s\r\n",
 			       vpninfo->disable_ipv6?"IPv4":"IPv6,IPv4");
-	openconnect_SSL_printf(vpninfo->https_ssl, "X-DTLS-Master-Secret: ");
+	openconnect_SSL_printf(vpninfo, "X-DTLS-Master-Secret: ");
 	for (i = 0; i < sizeof(vpninfo->dtls_secret); i++)
-		openconnect_SSL_printf(vpninfo->https_ssl, "%02X", vpninfo->dtls_secret[i]);
-	openconnect_SSL_printf(vpninfo->https_ssl, "\r\nX-DTLS-CipherSuite: %s\r\n\r\n",
+		openconnect_SSL_printf(vpninfo, "%02X", vpninfo->dtls_secret[i]);
+	openconnect_SSL_printf(vpninfo, "\r\nX-DTLS-CipherSuite: %s\r\n\r\n",
 			       vpninfo->dtls_ciphers?:"AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA");
 
-	if (openconnect_SSL_gets(vpninfo->https_ssl, buf, 65536) < 0) {
+	if ((i = openconnect_SSL_gets(vpninfo, buf, 65536) < 0)) {
+		if (i == -EINTR)
+			return i;
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Error fetching HTTPS response\n"));
 		if (!retried) {
@@ -150,7 +152,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 		if (!strncmp(buf, "HTTP/1.1 503 ", 13)) {
 			/* "Service Unavailable. Why? */
 			const char *reason = "<unknown>";
-			while ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
+			while ((i = openconnect_SSL_gets(vpninfo, buf, sizeof(buf)))) {
 				if (!strncmp(buf, "X-Reason: ", 10)) {
 					reason = buf + 10;
 					break;
@@ -174,9 +176,14 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	/* We may have advertised it, but we only do it if the server agrees */
 	vpninfo->deflate = 0;
 
-	while ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
+	while ((i = openconnect_SSL_gets(vpninfo, buf, sizeof(buf)))) {
 		struct vpn_option *new_option;
-		char *colon = strchr(buf, ':');
+		char *colon;
+
+		if (i < 0)
+			return i;
+
+		colon = strchr(buf, ':');
 		if (!colon)
 			continue;
 
@@ -352,10 +359,6 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	vpn_progress(vpninfo, PRG_INFO, _("CSTP connected. DPD %d, Keepalive %d\n"),
 		     vpninfo->ssl_times.dpd, vpninfo->ssl_times.keepalive);
 
-	BIO_set_nbio(SSL_get_rbio(vpninfo->https_ssl), 1);
-	BIO_set_nbio(SSL_get_wbio(vpninfo->https_ssl), 1);
-
-	fcntl(vpninfo->ssl_fd, F_SETFL, fcntl(vpninfo->ssl_fd, F_GETFL) | O_NONBLOCK);
 	if (vpninfo->select_nfds <= vpninfo->ssl_fd)
 		vpninfo->select_nfds = vpninfo->ssl_fd + 1;
 
@@ -394,11 +397,14 @@ int make_cstp_connection(struct openconnect_info *vpninfo)
 			if (!vpninfo->deflate_pkt) {
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Allocation of deflate buffer failed\n"));
+				inflateEnd(&vpninfo->inflate_strm);
+				deflateEnd(&vpninfo->deflate_strm);
 				vpninfo->deflate = 0;
+			} else {
+				memset(vpninfo->deflate_pkt, 0, sizeof(struct pkt));
+				memcpy(vpninfo->deflate_pkt->hdr, data_hdr, 8);
+				vpninfo->deflate_pkt->hdr[6] = AC_PKT_COMPRESSED;
 			}
-			memset(vpninfo->deflate_pkt, 0, sizeof(struct pkt));
-			memcpy(vpninfo->deflate_pkt->hdr, data_hdr, 8);
-			vpninfo->deflate_pkt->hdr[6] = AC_PKT_COMPRESSED;
 		}
 	}
 
@@ -413,11 +419,16 @@ int cstp_reconnect(struct openconnect_info *vpninfo)
 
 	openconnect_close_https(vpninfo);
 
-	/* It's already deflated in the old stream. Extremely
-	   non-trivial to reconstitute it; just throw it away */
-	if (vpninfo->current_ssl_pkt == vpninfo->deflate_pkt)
+	/* Requeue the original packet that was deflated */
+	if (vpninfo->current_ssl_pkt == vpninfo->deflate_pkt) {
 		vpninfo->current_ssl_pkt = NULL;
-
+		queue_packet(&vpninfo->outgoing_queue, vpninfo->pending_deflated_pkt);
+		vpninfo->pending_deflated_pkt = NULL;
+	}
+	if (vpninfo->deflate) {
+		inflateEnd(&vpninfo->inflate_strm);
+		deflateEnd(&vpninfo->deflate_strm);
+	}
 	timeout = vpninfo->reconnect_timeout;
 	interval = vpninfo->reconnect_interval;
 
@@ -435,7 +446,7 @@ int cstp_reconnect(struct openconnect_info *vpninfo)
 		if (interval > RECONNECT_INTERVAL_MAX)
 			interval = RECONNECT_INTERVAL_MAX;
 	}
-	script_reconnect(vpninfo);
+	script_config_tun(vpninfo, "reconnect");
 	return 0;
 }
 
@@ -611,7 +622,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 				return work_done;
 			default:
 				vpn_progress(vpninfo, PRG_ERR, _("SSL_write failed: %d\n"), ret);
-				report_ssl_errors(vpninfo);
+				openconnect_report_ssl_errors(vpninfo);
 				goto do_reconnect;
 			}
 		}
@@ -623,10 +634,11 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			return 1;
 		}
 		/* Don't free the 'special' packets */
-		if (vpninfo->current_ssl_pkt != vpninfo->deflate_pkt &&
-		    vpninfo->current_ssl_pkt != &dpd_pkt &&
-		    vpninfo->current_ssl_pkt != &dpd_resp_pkt &&
-		    vpninfo->current_ssl_pkt != &keepalive_pkt)
+		if (vpninfo->current_ssl_pkt == vpninfo->deflate_pkt)
+			free(vpninfo->pending_deflated_pkt);
+		else if (vpninfo->current_ssl_pkt != &dpd_pkt &&
+			 vpninfo->current_ssl_pkt != &dpd_resp_pkt &&
+			 vpninfo->current_ssl_pkt != &keepalive_pkt)
 			free(vpninfo->current_ssl_pkt);
 
 		vpninfo->current_ssl_pkt = NULL;
@@ -722,6 +734,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 				     _("Sending compressed data packet of %d bytes\n"),
 				     this->len);
 
+			vpninfo->pending_deflated_pkt = this;
 			vpninfo->current_ssl_pkt = vpninfo->deflate_pkt;
 		} else {
 		uncompr:
