@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2010 Intel Corporation.
+ * Copyright © 2008-2012 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -230,17 +230,15 @@ int connect_dtls_socket(struct openconnect_info *vpninfo)
 	}
 
 	/* Go Go Go! */
+	fcntl(dtls_fd, F_SETFL, fcntl(dtls_fd, F_GETFL) | O_NONBLOCK);
+
 	dtls_bio = BIO_new_socket(dtls_fd, BIO_NOCLOSE);
+	/* Set non-blocking */
+	BIO_set_nbio(dtls_bio, 1);
 	SSL_set_bio(dtls_ssl, dtls_bio, dtls_bio);
 
 	SSL_set_options(dtls_ssl, SSL_OP_CISCO_ANYCONNECT);
 
-	/* Set non-blocking */
-	BIO_set_nbio(SSL_get_rbio(dtls_ssl), 1);
-	BIO_set_nbio(SSL_get_wbio(dtls_ssl), 1);
-
-	fcntl(dtls_fd, F_SETFL, fcntl(dtls_fd, F_GETFL) | O_NONBLOCK);
-
 	vpninfo->new_dtls_fd = dtls_fd;
 	vpninfo->new_dtls_ssl = dtls_ssl;
 
@@ -300,19 +298,39 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 				     _("Your OpenSSL is older than the one you built against, so DTLS may fail!"));
 		}
 #elif defined (HAVE_DTLS1_STOP_TIMER)
+		/*
+		 * This works for any normal OpenSSL that supports
+		 * Cisco DTLS compatibility (0.9.8m to 1.0.0d inclusive,
+		 * and even later versions although it isn't needed there.
+		 */
 		dtls1_stop_timer(vpninfo->dtls_ssl);
-#else
-		/* Debian restricts visibility of dtls1_stop_timer()
-		   so do it manually. Thankfully this *should* work,
-		   from 0.9.8m to 1.0.0d inclusive, and we don't have
-		   to worry about future changes because we don't do
-		   this for 1.0.0e and above anyway */
+#elif defined (BIO_CTRL_DGRAM_SET_NEXT_TIMEOUT)
+		/*
+		 * Debian restricts visibility of dtls1_stop_timer()
+		 * so do it manually. This version also works on all
+		 * sane versions of OpenSSL:
+		 */
 		memset (&(vpninfo->dtls_ssl->d1->next_timeout), 0,
 			sizeof((vpninfo->dtls_ssl->d1->next_timeout)));
 		vpninfo->dtls_ssl->d1->timeout_duration = 1;
 		BIO_ctrl(SSL_get_rbio(vpninfo->dtls_ssl),
 			 BIO_CTRL_DGRAM_SET_NEXT_TIMEOUT, 0,
 			 &(vpninfo->dtls_ssl->d1->next_timeout));
+#elif defined (BIO_CTRL_DGRAM_SET_TIMEOUT)
+		/*
+		 * OK, here it gets more fun... this shoul handle the case
+		 * of older OpenSSL which has the Cisco DTLS compatibility
+		 * backported, but *not* the fix for RT#1922.
+		 */
+		BIO_ctrl(SSL_get_rbio(vpninfo->dtls_ssl),
+			 BIO_CTRL_DGRAM_SET_TIMEOUT, 0, NULL);
+#else
+		/*
+		 * And if they don't have any of the above, they probably
+		 * don't have RT#1829 fixed either, but that's OK because
+		 * that's the "fix" that *introduces* the timeout we're
+		 * trying to disable. So do nothing...
+		 */
 #endif
 		return 0;
 	}
@@ -325,7 +343,7 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 	}
 
 	vpn_progress(vpninfo, PRG_ERR, _("DTLS handshake failed: %d\n"), ret);
-	report_ssl_errors(vpninfo);
+	openconnect_report_ssl_errors(vpninfo);
 
 	/* Kill the new (failed) connection... */
 	SSL_free(vpninfo->new_dtls_ssl);
@@ -430,14 +448,29 @@ int setup_dtls(struct openconnect_info *vpninfo)
 	return 0;
 }
 
+static struct pkt *dtls_pkt;
+
 int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 {
-	unsigned char buf[2000];
-	int len;
 	int work_done = 0;
 	char magic_pkt;
 
-	while ( (len = SSL_read(vpninfo->dtls_ssl, buf, sizeof(buf))) > 0 ) {
+	while (1) {
+		int len = vpninfo->mtu;
+		unsigned char *buf;
+
+		if (!dtls_pkt) {
+			dtls_pkt = malloc(sizeof(struct pkt) + len);
+			if (!dtls_pkt) {
+				vpn_progress(vpninfo, PRG_ERR, "Allocation failed\n");
+				break;
+			}
+		}
+
+		buf = dtls_pkt->data - 1;
+		len = SSL_read(vpninfo->dtls_ssl, buf, len + 1);
+		if (len <= 0)
+			break;
 
 		vpn_progress(vpninfo, PRG_TRACE,
 			     _("Received DTLS packet 0x%02x of %d bytes\n"),
@@ -447,7 +480,9 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 
 		switch(buf[0]) {
 		case AC_PKT_DATA:
-			queue_new_packet(&vpninfo->incoming_queue, buf+1, len-1);
+			dtls_pkt->len = len - 1;
+			queue_packet(&vpninfo->incoming_queue, dtls_pkt);
+			dtls_pkt = NULL;
 			work_done = 1;
 			break;
 
@@ -563,7 +598,7 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("DTLS got write error %d. Falling back to SSL\n"),
 					     ret);
-				report_ssl_errors(vpninfo);
+				openconnect_report_ssl_errors(vpninfo);
 				dtls_restart(vpninfo);
 				vpninfo->outgoing_queue = this;
 				vpninfo->outgoing_qlen++;
