@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2010 Intel Corporation.
+ * Copyright © 2008-2012 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
@@ -39,7 +39,10 @@
 
 #include "openconnect-internal.h"
 
-static int proxy_write(int fd, unsigned char *buf, size_t len);
+static int proxy_write(struct openconnect_info *vpninfo, int fd,
+		       unsigned char *buf, size_t len);
+static int proxy_read(struct openconnect_info *vpninfo, int fd,
+		      unsigned char *buf, size_t len);
 
 #define MAX_BUF_LEN 131072
 /*
@@ -112,7 +115,7 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 	int i;
 
  cont:
-	if (openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)) < 0) {
+	if (openconnect_SSL_gets(vpninfo, buf, sizeof(buf)) < 0) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Error fetching HTTPS response\n"));
 		return -EINVAL;
@@ -131,7 +134,7 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		     _("Got HTTP response: %s\n"), buf);
 
 	/* Eat headers... */
-	while ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
+	while ((i = openconnect_SSL_gets(vpninfo, buf, sizeof(buf)))) {
 		char *colon;
 
 		if (i < 0) {
@@ -247,7 +250,7 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		if (!body)
 			return -ENOMEM;
 		while (done < bodylen) {
-			i = SSL_read(vpninfo->https_ssl, body + done, bodylen - done);
+			i = openconnect_SSL_read(vpninfo, body + done, bodylen - done);
 			if (i < 0) {
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Error reading HTTP response body\n"));
@@ -258,13 +261,13 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 		}
 	} else if (bodylen == BODY_CHUNKED) {
 		/* ... else, chunked */
-		while ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
+		while ((i = openconnect_SSL_gets(vpninfo, buf, sizeof(buf)))) {
 			int chunklen, lastchunk = 0;
 
 			if (i < 0) {
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Error fetching chunk header\n"));
-				exit(1);
+				return i;
 			}
 			chunklen = strtol(buf, NULL, 16);
 			if (!chunklen) {
@@ -275,7 +278,7 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 			if (!body)
 				return -ENOMEM;
 			while (chunklen) {
-				i = SSL_read(vpninfo->https_ssl, body + done, chunklen);
+				i = openconnect_SSL_read(vpninfo, body + done, chunklen);
 				if (i < 0) {
 					vpn_progress(vpninfo, PRG_ERR,
 						     _("Error reading HTTP response body\n"));
@@ -286,7 +289,7 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 				done += i;
 			}
 		skip:
-			if ((i = openconnect_SSL_gets(vpninfo->https_ssl, buf, sizeof(buf)))) {
+			if ((i = openconnect_SSL_gets(vpninfo, buf, sizeof(buf)))) {
 				if (i < 0) {
 					vpn_progress(vpninfo, PRG_ERR,
 						     _("Error fetching HTTP response body\n"));
@@ -314,14 +317,21 @@ static int process_http_response(struct openconnect_info *vpninfo, int *result,
 			body = realloc(body, done + 16384);
 			if (!body)
 				return -ENOMEM;
-			i = SSL_read(vpninfo->https_ssl, body + done, 16384);
-			if (i <= 0) {
+			i = openconnect_SSL_read(vpninfo, body + done, 16384);
+			if (i > 0) {
+				/* Got more data */
+				done += i;
+			} else if (i < 0) {
+				/* Error */
+				free(body);
+				return i;
+			} else {
+				/* Connection closed. Reduce allocation to just what we need */
 				body = realloc(body, done + 1);
 				if (!body)
 					return -ENOMEM;
 				break;
                         }
-			done += i;
 		}
 	}
 
@@ -423,7 +433,7 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 		return err;
 	}
 
-	ret = proxy_write(fd, (void *)buf, buflen);
+	ret = proxy_write(vpninfo, fd, (void *)buf, buflen);
 	if (ret) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to write temporary CSD script file: %s\n"),
@@ -445,13 +455,13 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 			struct passwd *pw;
 
 			if (setuid(vpninfo->uid_csd)) {
-				fprintf(stderr, _("Failed to set uid %d\n"),
-					vpninfo->uid_csd);
+				fprintf(stderr, _("Failed to set uid %ld\n"),
+					(long)vpninfo->uid_csd);
 				exit(1);
 			}
 			if (!(pw = getpwuid(vpninfo->uid_csd))) {
-				fprintf(stderr, _("Invalid user uid=%d\n"),
-					vpninfo->uid_csd);
+				fprintf(stderr, _("Invalid user uid=%ld\n"),
+					(long)vpninfo->uid_csd);
 				exit(1);
 			}
 			setenv("HOME", pw->pw_dir, 1);
@@ -525,8 +535,8 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 	return 0;
 }
 
-#ifdef __sun__
-static char *local_strcasestr(const char *haystack, const char *needle)
+#ifndef HAVE_STRCASESTR
+static char *openconnect__strcasestr(const char *haystack, const char *needle)
 {
 	int hlen = strlen(haystack);
 	int nlen = strlen(needle);
@@ -543,9 +553,10 @@ static char *local_strcasestr(const char *haystack, const char *needle)
 	}
 	return NULL;
 }
-#define strcasestr local_strcasestr
+#define strcasestr openconnect__strcasestr
 #endif
 
+
 int internal_parse_url(char *url, char **res_proto, char **res_host,
 		       int *res_port, char **res_path, int default_port)
 {
@@ -602,6 +613,8 @@ int internal_parse_url(char *url, char **res_proto, char **res_host,
 		*res_path = (path && *path) ? strdup(path) : NULL;
 
 	/* Undo the damage we did to the original string */
+	if (port_str)
+		*(port_str) = ':';
 	if (path)
 		*(path - 1) = '/';
 	if (proto)
@@ -676,12 +689,20 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 			     method, vpninfo->hostname, vpninfo->port,
 			     vpninfo->urlpath ?: "");
 
-	SSL_write(vpninfo->https_ssl, buf, strlen(buf));
+	result = openconnect_SSL_write(vpninfo, buf, strlen(buf));
+	if (result < 0)
+		return result;
+
+	/* Remember the peer's SSL certificate; it may disconnect during
+	   the response and then we wouldn't be able to find it */
+	if (vpninfo->peer_cert)
+		X509_free(vpninfo->peer_cert);
+	vpninfo->peer_cert = SSL_get_peer_certificate(vpninfo->https_ssl);
 
 	buflen = process_http_response(vpninfo, &result, NULL, &form_buf);
 	if (buflen < 0) {
 		/* We'll already have complained about whatever offended us */
-		exit(1);
+		return buflen;
 	}
 
 	if (result != 200 && vpninfo->redirect_url) {
@@ -701,6 +722,7 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 					     _("Failed to parse redirected URL '%s': %s\n"),
 					     vpninfo->redirect_url, strerror(-ret));
 				free(vpninfo->redirect_url);
+				vpninfo->redirect_url = NULL;
 				free(form_buf);
 				return ret;
 			}
@@ -735,6 +757,14 @@ int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 			vpninfo->redirect_url = NULL;
 
 			goto retry;
+		} else if (strstr(vpninfo->redirect_url, "://")) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Cannot follow redirection to non-https URL '%s'\n"),
+				     vpninfo->redirect_url);
+			free(vpninfo->redirect_url);
+			vpninfo->redirect_url = NULL;
+			free(form_buf);
+			return -EINVAL;
 		} else if (vpninfo->redirect_url[0] == '/') {
 			/* Absolute redirect within same host */
 			free(vpninfo->urlpath);
@@ -856,13 +886,14 @@ char *openconnect_create_useragent(const char *base)
 {
 	char *uagent;
 
-	if (asprintf(&uagent, "%s %s", base, openconnect_version) < 0)
+	if (asprintf(&uagent, "%s %s", base, openconnect_version_str) < 0)
 		return NULL;
 
 	return uagent;
 }
 
-static int proxy_gets(int fd, char *buf, size_t len)
+static int proxy_gets(struct openconnect_info *vpninfo, int fd,
+		      char *buf, size_t len)
 {
 	int i = 0;
 	int ret;
@@ -870,7 +901,7 @@ static int proxy_gets(int fd, char *buf, size_t len)
 	if (len < 2)
 		return -EINVAL;
 
-	while ( (ret = read(fd, buf + i, 1)) == 1) {
+	while ( (ret = proxy_read(vpninfo, fd, (void *)(buf + i), 1)) == 0) {
 		if (buf[i] == '\n') {
 			buf[i] = 0;
 			if (i && buf[i-1] == '\r') {
@@ -886,19 +917,39 @@ static int proxy_gets(int fd, char *buf, size_t len)
 			return i;
 		}
 	}
-	if (ret < 0)
-		ret = -errno;
-
 	buf[i] = 0;
 	return i ?: ret;
 }
 
-static int proxy_write(int fd, unsigned char *buf, size_t len)
+static int proxy_write(struct openconnect_info *vpninfo, int fd,
+		       unsigned char *buf, size_t len)
 {
 	size_t count;
-	
+
 	for (count = 0; count < len; ) {
-		int i = write(fd, buf + count, len - count);
+		fd_set rd_set, wr_set;
+		int maxfd = fd;
+		int i;
+
+		FD_ZERO(&wr_set);
+		FD_ZERO(&rd_set);
+		FD_SET(fd, &wr_set);
+		if (vpninfo->cancel_fd != -1) {
+			FD_SET(vpninfo->cancel_fd, &rd_set);
+			if (vpninfo->cancel_fd > fd)
+				maxfd = vpninfo->cancel_fd;
+		}
+
+		select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+		if (vpninfo->cancel_fd != -1 &&
+		    FD_ISSET(vpninfo->cancel_fd, &rd_set))
+			return -EINTR;
+
+		/* Not that this should ever be able to happen... */
+		if (!FD_ISSET(fd, &wr_set))
+			continue;
+
+		i = write(fd, buf + count, len - count);
 		if (i < 0)
 			return -errno;
 
@@ -907,12 +958,34 @@ static int proxy_write(int fd, unsigned char *buf, size_t len)
 	return 0;
 }
 
-static int proxy_read(int fd, unsigned char *buf, size_t len)
+static int proxy_read(struct openconnect_info *vpninfo, int fd,
+		      unsigned char *buf, size_t len)
 {
 	size_t count;
 
 	for (count = 0; count < len; ) {
-		int i = read(fd, buf + count, len - count);
+		fd_set rd_set;
+		int maxfd = fd;
+		int i;
+
+		FD_ZERO(&rd_set);
+		FD_SET(fd, &rd_set);
+		if (vpninfo->cancel_fd != -1) {
+			FD_SET(vpninfo->cancel_fd, &rd_set);
+			if (vpninfo->cancel_fd > fd)
+				maxfd = vpninfo->cancel_fd;
+		}
+
+		select(maxfd + 1, &rd_set, NULL, NULL, NULL);
+		if (vpninfo->cancel_fd != -1 &&
+		    FD_ISSET(vpninfo->cancel_fd, &rd_set))
+			return -EINTR;
+
+		/* Not that this should ever be able to happen... */
+		if (!FD_ISSET(fd, &rd_set))
+			continue;
+
+		i = read(fd, buf + count, len - count);
 		if (i < 0)
 			return -errno;
 
@@ -942,14 +1015,14 @@ static int process_socks_proxy(struct openconnect_info *vpninfo, int ssl_sock)
 	buf[1] = 1; /* # auth methods */
 	buf[2] = 0; /* No auth supported */
 
-	if ((i = proxy_write(ssl_sock, buf, 3))) {
+	if ((i = proxy_write(vpninfo, ssl_sock, buf, 3))) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Error writing auth request to SOCKS proxy: %s\n"),
 			     strerror(-i));
 		return i;
 	}
 	
-	if ((i = proxy_read(ssl_sock, buf, 2))) {
+	if ((i = proxy_read(vpninfo, ssl_sock, buf, 2))) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Error reading auth response from SOCKS proxy: %s\n"),
 			     strerror(-i));
@@ -988,7 +1061,7 @@ static int process_socks_proxy(struct openconnect_info *vpninfo, int ssl_sock)
 	buf[i++] = vpninfo->port >> 8;
 	buf[i++] = vpninfo->port & 0xff;
 
-	if ((i = proxy_write(ssl_sock, buf, i))) {
+	if ((i = proxy_write(vpninfo, ssl_sock, buf, i))) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Error writing connect request to SOCKS proxy: %s\n"),
 			     strerror(-i));
@@ -996,7 +1069,7 @@ static int process_socks_proxy(struct openconnect_info *vpninfo, int ssl_sock)
 	}
 	/* Read 5 bytes -- up to and including the first byte of the returned
 	   address (which might be the length byte of a domain name) */
-	if ((i = proxy_read(ssl_sock, buf, 5))) {
+	if ((i = proxy_read(vpninfo, ssl_sock, buf, 5))) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Error reading connect response from SOCKS proxy: %s\n"),
 			     strerror(-i));
@@ -1029,7 +1102,7 @@ static int process_socks_proxy(struct openconnect_info *vpninfo, int ssl_sock)
 		return -EIO;
 	}
 
-	if ((i = proxy_read(ssl_sock, buf, i))) {
+	if ((i = proxy_read(vpninfo, ssl_sock, buf, i))) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Error reading connect response from SOCKS proxy: %s\n"),
 			     strerror(-i));
@@ -1055,15 +1128,15 @@ static int process_http_proxy(struct openconnect_info *vpninfo, int ssl_sock)
 		     _("Requesting HTTP proxy connection to %s:%d\n"),
 		     vpninfo->hostname, vpninfo->port);
 
-	if (proxy_write(ssl_sock, (unsigned char *)buf, strlen(buf))) {
-		result = -errno;
+	result = proxy_write(vpninfo, ssl_sock, (unsigned char *)buf, strlen(buf));
+	if (result) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Sending proxy request failed: %s\n"),
-			     strerror(errno));
+			     strerror(-result));
 		return result;
 	}
 
-	if (proxy_gets(ssl_sock, buf, sizeof(buf)) < 0) {
+	if (proxy_gets(vpninfo, ssl_sock, buf, sizeof(buf)) < 0) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Error fetching proxy response\n"));
 		return -EIO;
@@ -1082,7 +1155,7 @@ static int process_http_proxy(struct openconnect_info *vpninfo, int ssl_sock)
 		return -EIO;
 	}
 
-	while ((buflen = proxy_gets(ssl_sock, buf, sizeof(buf)))) {
+	while ((buflen = proxy_gets(vpninfo, ssl_sock, buf, sizeof(buf)))) {
 		if (buflen < 0) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Failed to read proxy response\n"));
