@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2011 Intel Corporation.
+ * Copyright © 2008-2012 Intel Corporation.
  *
  * Authors: David Woodhouse <dwmw2@infradead.org>
  *
@@ -22,6 +22,9 @@
  *   Boston, MA 02110-1301 USA
  */
 
+#include <string.h>
+#include <errno.h>
+
 #include "openconnect-internal.h"
 
 struct openconnect_info *openconnect_vpninfo_new_with_cbdata (char *useragent,
@@ -42,6 +45,7 @@ struct openconnect_info *openconnect_vpninfo_new_with_cbdata (char *useragent,
 	vpninfo->process_auth_form = process_auth_form;
 	vpninfo->progress = progress;
 	vpninfo->cbdata = privdata?:vpninfo;
+	vpninfo->cancel_fd = -1;
 
 #ifdef ENABLE_NLS
 	bindtextdomain("openconnect", LOCALEDIR);
@@ -95,6 +99,8 @@ void openconnect_vpninfo_free (struct openconnect_info *vpninfo)
 	if (vpninfo->cert != vpninfo->sslkey)
 		free((void *)vpninfo->sslkey);
 	free((void *)vpninfo->cert);
+	if (vpninfo->peer_cert)
+		X509_free(vpninfo->peer_cert);
 	/* No need to free deflate streams; they weren't initialised */
 	free(vpninfo);
 }
@@ -150,7 +156,7 @@ void openconnect_set_client_cert (struct openconnect_info *vpninfo, char *cert,
 
 struct x509_st *openconnect_get_peer_cert (struct openconnect_info *vpninfo)
 {
-	return SSL_get_peer_certificate(vpninfo->https_ssl);
+	return vpninfo->peer_cert;
 }
 
 int openconnect_get_port (struct openconnect_info *vpninfo)
@@ -186,13 +192,35 @@ void openconnect_reset_ssl (struct openconnect_info *vpninfo)
 
 int openconnect_parse_url (struct openconnect_info *vpninfo, char *url)
 {
+	char *scheme = NULL;
+	int ret;
+
 	if (vpninfo->peer_addr) {
 		free(vpninfo->peer_addr);
 		vpninfo->peer_addr = NULL;
 	}
 
-	return internal_parse_url (url, NULL, &vpninfo->hostname,
-				   &vpninfo->port, &vpninfo->urlpath, 443);
+	free(vpninfo->hostname);
+	vpninfo->hostname = NULL;
+	free(vpninfo->urlpath);
+	vpninfo->urlpath = NULL;
+
+	ret = internal_parse_url (url, &scheme, &vpninfo->hostname,
+				  &vpninfo->port, &vpninfo->urlpath, 443);
+
+	if (ret) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to parse server URL '%s'\n"),
+			     url);
+		return ret;
+	}
+	if (scheme && strcmp(scheme, "https")) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Only https:// permitted for server URL\n"));
+		ret = -EINVAL;
+	}
+	free(scheme);
+	return ret;
 }
 
 void openconnect_set_cert_expiry_warning (struct openconnect_info *vpninfo,
@@ -201,7 +229,12 @@ void openconnect_set_cert_expiry_warning (struct openconnect_info *vpninfo,
 	vpninfo->cert_expire_warning = seconds;
 }
 
+void openconnect_set_cancel_fd (struct openconnect_info *vpninfo, int fd)
+{
+	vpninfo->cancel_fd = fd;
+}
+
 const char *openconnect_get_version (void)
 {
-	return openconnect_version;
+	return openconnect_version_str;
 }
