@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2011 Intel Corporation.
+ * Copyright © 2008-2012 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
@@ -23,6 +23,11 @@
  *   Boston, MA 02110-1301 USA
  */
 
+#ifdef HAVE_GETLINE
+/* Various BSD systems require this for getline() to be visible */
+#define _WITH_GETLINE
+#endif
+
 #include <stdio.h>
 #ifdef ANDROID
 #include <android/log.h>
@@ -62,9 +67,9 @@ static int validate_peer_cert(void *_vpninfo,
 
 /* A sanity check that the openconnect executable is running against a
    library of the same version */
-#define openconnect_version openconnect_binary_version
+#define openconnect_version_str openconnect_binary_version
 #include "version.c"
-#undef openconnect_version
+#undef openconnect_version_str
 
 int verbose = PRG_INFO;
 int background;
@@ -75,6 +80,7 @@ int non_inter;
 enum {
 	OPT_AUTHGROUP = 0x100,
 	OPT_CAFILE,
+	OPT_CONFIGFILE,
 	OPT_COOKIEONLY,
 	OPT_COOKIE_ON_STDIN,
 	OPT_CSD_USER,
@@ -98,61 +104,79 @@ enum {
 	OPT_NON_INTER,
 };
 
+#ifdef __sun__
+/*
+ * The 'name' field in Solaris 'struct option' lacks the 'const', and causes
+ * lots of warnings unless we cast it... https://www.illumos.org/issues/1881
+*/
+#define OPTION(name, arg, abbrev) {(char *)name, arg, NULL, abbrev}
+#else
+#define OPTION(name, arg, abbrev) {name, arg, NULL, abbrev}
+#endif
+
 static struct option long_options[] = {
-	{"background", 0, 0, 'b'},
-	{"pid-file", 1, 0, OPT_PIDFILE},
-	{"certificate", 1, 0, 'c'},
-	{"sslkey", 1, 0, 'k'},
-	{"cookie", 1, 0, 'C'},
-	{"deflate", 0, 0, 'd'},
-	{"no-deflate", 0, 0, 'D'},
-	{"cert-expire-warning", 1, 0, 'e'},
-	{"usergroup", 1, 0, 'g'},
-	{"help", 0, 0, 'h'},
-	{"interface", 1, 0, 'i'},
-	{"mtu", 1, 0, 'm'},
-	{"setuid", 1, 0, 'U'},
-	{"script", 1, 0, 's'},
-	{"script-tun", 0, 0, 'S'},
-	{"syslog", 0, 0, 'l'},
-	{"key-type", 1, 0, 'K'},
-	{"key-password", 1, 0, 'p'},
-	{"proxy", 1, 0, 'P'},
-	{"user", 1, 0, 'u'},
-	{"verbose", 0, 0, 'v'},
-	{"version", 0, 0, 'V'},
-	{"cafile", 1, 0, OPT_CAFILE},
-	{"no-dtls", 0, 0, OPT_NO_DTLS},
-	{"cookieonly", 0, 0, OPT_COOKIEONLY},
-	{"printcookie", 0, 0, OPT_PRINTCOOKIE},
-	{"quiet", 0, 0, 'q'},
-	{"queue-len", 1, 0, 'Q'},
-	{"xmlconfig", 1, 0, 'x'},
-	{"cookie-on-stdin", 0, 0, OPT_COOKIE_ON_STDIN},
-	{"passwd-on-stdin", 0, 0, OPT_PASSWORD_ON_STDIN},
-	{"no-passwd", 0, 0, OPT_NO_PASSWD},
-	{"reconnect-timeout", 1, 0, OPT_RECONNECT_TIMEOUT},
-	{"dtls-ciphers", 1, 0, OPT_DTLS_CIPHERS},
-	{"authgroup", 1, 0, OPT_AUTHGROUP},
-	{"servercert", 1, 0, OPT_SERVERCERT},
-	{"key-password-from-fsid", 0, 0, OPT_KEY_PASSWORD_FROM_FSID},
-	{"useragent", 1, 0, OPT_USERAGENT},
-	{"csd-user", 1, 0, OPT_CSD_USER},
-	{"csd-wrapper", 1, 0, OPT_CSD_WRAPPER},
-	{"disable-ipv6", 0, 0, OPT_DISABLE_IPV6},
-	{"no-proxy", 0, 0, OPT_NO_PROXY},
-	{"libproxy", 0, 0, OPT_LIBPROXY},
-	{"no-http-keepalive", 0, 0, OPT_NO_HTTP_KEEPALIVE},
-	{"no-cert-check", 0, 0, OPT_NO_CERT_CHECK},
-	{"force-dpd", 1, 0, OPT_FORCE_DPD},
-	{"non-inter", 0, 0, OPT_NON_INTER},
-	{NULL, 0, 0, 0},
+	OPTION("background", 0, 'b'),
+	OPTION("pid-file", 1, OPT_PIDFILE),
+	OPTION("certificate", 1, 'c'),
+	OPTION("sslkey", 1, 'k'),
+	OPTION("cookie", 1, 'C'),
+	OPTION("deflate", 0, 'd'),
+	OPTION("no-deflate", 0, 'D'),
+	OPTION("cert-expire-warning", 1, 'e'),
+	OPTION("usergroup", 1, 'g'),
+	OPTION("help", 0, 'h'),
+	OPTION("interface", 1, 'i'),
+	OPTION("mtu", 1, 'm'),
+	OPTION("setuid", 1, 'U'),
+	OPTION("script", 1, 's'),
+	OPTION("script-tun", 0, 'S'),
+	OPTION("syslog", 0, 'l'),
+	OPTION("key-type", 1, 'K'),
+	OPTION("key-password", 1, 'p'),
+	OPTION("proxy", 1, 'P'),
+	OPTION("user", 1, 'u'),
+	OPTION("verbose", 0, 'v'),
+	OPTION("version", 0, 'V'),
+	OPTION("cafile", 1, OPT_CAFILE),
+	OPTION("config", 1, OPT_CONFIGFILE),
+	OPTION("no-dtls", 0, OPT_NO_DTLS),
+	OPTION("cookieonly", 0, OPT_COOKIEONLY),
+	OPTION("printcookie", 0, OPT_PRINTCOOKIE),
+	OPTION("quiet", 0, 'q'),
+	OPTION("queue-len", 1, 'Q'),
+	OPTION("xmlconfig", 1, 'x'),
+	OPTION("cookie-on-stdin", 0, OPT_COOKIE_ON_STDIN),
+	OPTION("passwd-on-stdin", 0, OPT_PASSWORD_ON_STDIN),
+	OPTION("no-passwd", 0, OPT_NO_PASSWD),
+	OPTION("reconnect-timeout", 1, OPT_RECONNECT_TIMEOUT),
+	OPTION("dtls-ciphers", 1, OPT_DTLS_CIPHERS),
+	OPTION("authgroup", 1, OPT_AUTHGROUP),
+	OPTION("servercert", 1, OPT_SERVERCERT),
+	OPTION("key-password-from-fsid", 0, OPT_KEY_PASSWORD_FROM_FSID),
+	OPTION("useragent", 1, OPT_USERAGENT),
+	OPTION("csd-user", 1, OPT_CSD_USER),
+	OPTION("csd-wrapper", 1, OPT_CSD_WRAPPER),
+	OPTION("disable-ipv6", 0, OPT_DISABLE_IPV6),
+	OPTION("no-proxy", 0, OPT_NO_PROXY),
+	OPTION("libproxy", 0, OPT_LIBPROXY),
+	OPTION("no-http-keepalive", 0, OPT_NO_HTTP_KEEPALIVE),
+	OPTION("no-cert-check", 0, OPT_NO_CERT_CHECK),
+	OPTION("force-dpd", 1, OPT_FORCE_DPD),
+	OPTION("non-inter", 0, OPT_NON_INTER),
+	OPTION(NULL, 0, 0)
 };
 
+static void helpmessage(void)
+{
+	printf(_("For assistance with OpenConnect, please see the web page at\n"
+		 "  http://www.infradead.org/openconnect/mail.html\n"));
+}
+
 static void usage(void)
 {
 	printf(_("Usage:  openconnect [options] <server>\n"));
-	printf(_("Open client for Cisco AnyConnect VPN, version %s\n\n"), openconnect_version);
+	printf(_("Open client for Cisco AnyConnect VPN, version %s\n\n"), openconnect_version_str);
+	printf("      --config=CONFIGFILE         %s\n", _("Read options from config file"));
 	printf("  -b, --background                %s\n", _("Continue in background after startup"));
 	printf("      --pid-file=PIDFILE          %s\n", _("Write the daemons pid to this file"));
 	printf("  -c, --certificate=CERT          %s\n", _("Use SSL client certificate CERT"));
@@ -183,6 +207,7 @@ static void usage(void)
 	printf("  -q, --quiet                     %s\n", _("Less output"));
 	printf("  -Q, --queue-len=LEN             %s\n", _("Set packet queue limit to LEN pkts"));
 	printf("  -s, --script=SCRIPT             %s\n", _("Shell command line for using a vpnc-compatible config script"));
+	printf("                                  %s: \"%s\"\n", _("default"), DEFAULT_VPNCSCRIPT);
 	printf("  -S, --script-tun                %s\n", _("Pass traffic to 'script' program, not tun"));
 	printf("  -u, --user=NAME                 %s\n", _("Set login username"));
 	printf("  -V, --version                   %s\n", _("Report version number"));
@@ -203,6 +228,9 @@ static void usage(void)
 	printf("      --reconnect-timeout         %s\n", _("Connection retry timeout in seconds"));
 	printf("      --servercert=FINGERPRINT    %s\n", _("Server's certificate SHA1 fingerprint"));
 	printf("      --useragent=STRING          %s\n", _("HTTP header User-Agent: field"));
+	printf("\n");
+
+	helpmessage();
 	exit(1);
 }
 
@@ -232,6 +260,119 @@ static void handle_sigusr(int sig)
 		verbose = PRG_INFO;
 }
 
+static FILE *config_file = NULL;
+static int config_line_num = 0;
+
+#define keep_config_arg() (config_arg)
+
+static int next_option(int argc, char **argv, char **config_arg)
+{
+	/* These get re-used */
+	static char *line_buf = NULL;
+	static size_t line_size = 0;
+
+	ssize_t llen;
+	int opt, optlen;
+	struct option *this;
+	char *line;
+	int ate_equals = 0;
+
+ next:
+	if (!config_file) {
+		opt = getopt_long(argc, argv,
+				  "bC:c:e:Ddg:hi:k:K:lpP:Q:qSs:U:u:Vvx:",
+				  long_options, NULL);
+
+		*config_arg = optarg;
+		return opt;
+	}
+
+	llen = getline(&line_buf, &line_size, config_file);
+	if (llen < 0) {
+		if (feof(config_file)) {
+			fclose(config_file);
+			config_file = NULL;
+			goto next;
+		}
+		fprintf(stderr, _("Failed to get line from config file: %s\n"),
+			strerror(errno));
+		exit(1);
+	}
+	line = line_buf;
+
+	/* Strip the trailing newline (coping with DOS newlines) */
+	if (llen && line[llen-1] == '\n')
+		line[--llen] = 0;
+	if (llen && line[llen-1] == '\r')
+		line[--llen] = 0;
+
+	/* Skip and leading whitespace */
+	while (line[0] == ' ' || line[0] == '\t' || line[0] == '\r')
+		line++;
+
+	/* Ignore comments and empty lines */
+	if (!line[0] || line[0] == '#') {
+		config_line_num++;
+		goto next;
+	}
+
+	/* Try to match on a known option... naïvely. This could be improved. */
+	for (this = long_options; this->name; this++) {
+		optlen = strlen(this->name);
+		/* If the option isn't followed by whitespace or NUL, or
+		   perhaps an equals sign if the option takes an argument,
+		   then it's not a match */
+		if (!strncmp(this->name, line, optlen) &&
+		    (!line[optlen] || line[optlen] == ' ' || line[optlen] == '\t' ||
+		     line[optlen] == '='))
+		    break;
+	}
+	if (!this->name) {
+		char *l;
+
+		for (l = line; *l && *l != ' ' && *l != '\t'; l++)
+			;
+
+		*l = 0;
+		fprintf(stderr, _("Unrecognised option at line %d: '%s'\n"),
+			config_line_num, line);
+		return '?';
+	}
+	line += optlen;
+	while (*line == ' ' || *line == '\t' ||
+	       (*line == '=' && this->has_arg && !ate_equals && ++ate_equals))
+		line++;
+
+	if (!this->has_arg && *line) {
+		fprintf(stderr, _("Option '%s' does not take an argument at line %d\n"),
+			this->name, config_line_num);
+		return '?';
+	} else if (this->has_arg && !*line) {
+		fprintf(stderr, _("Option '%s' requires an argument at line %d\n"),
+			this->name, config_line_num);
+		return '?';
+	}
+
+	config_line_num++;
+	*config_arg = line;
+	return this->val;
+
+}
+
+static int null_ui_open(UI *ui)
+{
+	fprintf(stderr, _("Cannot ask for user input while in non-interactive mode\n"));
+	fflush(stderr);
+	return 0;
+}
+
+static void disable_openssl_ui()
+{
+	UI_METHOD *ui_method = UI_create_method((char *)"OpenConnect non-interactive UI");
+	UI_method_set_opener(ui_method, null_ui_open);
+	UI_set_default_method(ui_method);
+}
+
 int main(int argc, char **argv)
 {
 	struct openconnect_info *vpninfo;
@@ -239,22 +380,24 @@ int main(int argc, char **argv)
 	struct sigaction sa;
 	int cookieonly = 0;
 	int use_syslog = 0;
+	char *urlpath = NULL;
 	char *proxy = getenv("https_proxy");
 	int autoproxy = 0;
 	uid_t uid = getuid();
 	int opt;
 	char *pidfile = NULL;
 	FILE *fp = NULL;
+	char *config_arg;
 
 #ifdef ENABLE_NLS
 	bindtextdomain("openconnect", LOCALEDIR);
 	setlocale(LC_ALL, "");
 #endif
 
-	if (strcmp(openconnect_version, openconnect_binary_version)) {
+	if (strcmp(openconnect_version_str, openconnect_binary_version)) {
 		fprintf(stderr, _("WARNING: This version of openconnect is %s but\n"
 				  "         the libopenconnect library is %s\n"),
-			openconnect_binary_version, openconnect_version);
+			openconnect_binary_version, openconnect_version_str);
 	}
 			
 	openconnect_init_openssl();
@@ -280,26 +423,42 @@ int main(int argc, char **argv)
 	vpninfo->validate_peer_cert = validate_peer_cert;
 	vpninfo->cbdata = vpninfo;
 	vpninfo->cert_expire_warning = 60 * 86400;
+	vpninfo->vpnc_script = DEFAULT_VPNCSCRIPT;
+	vpninfo->cancel_fd = -1;
 
 	if (!uname(&utsbuf))
 		vpninfo->localname = utsbuf.nodename;
 	else
 		vpninfo->localname = "localhost";
 
-	while ((opt = getopt_long(argc, argv, "bC:c:e:Ddg:hi:k:K:lpP:Q:qSs:U:u:Vvx:",
-				  long_options, NULL))) {
+	while ((opt = next_option(argc, argv, &config_arg))) {
+
 		if (opt < 0)
 			break;
 
 		switch (opt) {
+		case OPT_CONFIGFILE:
+			if (config_file) {
+				fprintf(stderr, _("Cannot use 'config' option inside config file\n"));
+				exit(1);
+			}
+			config_file = fopen(config_arg, "r");
+			if (!config_file) {
+				fprintf(stderr, _("Cannot open config file '%s': %s\n"),
+					config_arg, strerror(errno));
+				exit(1);
+			}
+			config_line_num = 1;
+			/* The next option will come from the file... */
+			break;
 		case OPT_CAFILE:
-			vpninfo->cafile = optarg;
+			vpninfo->cafile = keep_config_arg();
 			break;
 		case OPT_PIDFILE:
-			pidfile = optarg;
+			pidfile = keep_config_arg();
 			break;
 		case OPT_SERVERCERT:
-			vpninfo->servercert = optarg;
+			vpninfo->servercert = keep_config_arg();
 			break;
 		case OPT_NO_DTLS:
 			vpninfo->dtls_attempt_period = 0;
@@ -325,43 +484,43 @@ int main(int argc, char **argv)
 			break;
 		case OPT_NON_INTER:
 			non_inter = 1;
-			vpninfo->nopasswd = 1;
+			disable_openssl_ui();
 			break;
 		case OPT_RECONNECT_TIMEOUT:
-			vpninfo->reconnect_timeout = atoi(optarg);
+			vpninfo->reconnect_timeout = atoi(config_arg);
 			break;
 		case OPT_DTLS_CIPHERS:
-			vpninfo->dtls_ciphers = optarg;
+			vpninfo->dtls_ciphers = keep_config_arg();
 			break;
 		case OPT_AUTHGROUP:
-			vpninfo->authgroup = optarg;
+			vpninfo->authgroup = keep_config_arg();
 			break;
 		case 'b':
 			background = 1;
 			break;
 		case 'C':
-			vpninfo->cookie = optarg;
+			vpninfo->cookie = keep_config_arg();
 			break;
 		case 'c':
-			vpninfo->cert = optarg;
+			vpninfo->cert = keep_config_arg();
 			break;
 		case 'e':
-			vpninfo->cert_expire_warning = 86400 * atoi(optarg);
+			vpninfo->cert_expire_warning = 86400 * atoi(config_arg);
 			break;
 		case 'k':
-			vpninfo->sslkey = optarg;
+			vpninfo->sslkey = keep_config_arg();
 			break;
 		case 'K':
-			if (!strcasecmp(optarg, "PKCS#12") ||
-			    !strcasecmp(optarg, "PKCS12")) {
+			if (!strcasecmp(config_arg, "PKCS#12") ||
+			    !strcasecmp(config_arg, "PKCS12")) {
 				vpninfo->cert_type = CERT_TYPE_PKCS12;
-			} else if (!strcasecmp(optarg, "TPM")) {
+			} else if (!strcasecmp(config_arg, "TPM")) {
 				vpninfo->cert_type = CERT_TYPE_TPM;
-			} else if (!strcasecmp(optarg, "PEM")) {
+			} else if (!strcasecmp(config_arg, "PEM")) {
 				vpninfo->cert_type = CERT_TYPE_PEM;
 			} else {
 				fprintf(stderr, _("Unknown certificate type '%s'\n"),
-					optarg);
+					config_arg);
 				usage();
 			}
 		case 'd':
@@ -371,29 +530,29 @@ int main(int argc, char **argv)
 			vpninfo->deflate = 0;
 			break;
 		case 'g':
-			free(vpninfo->urlpath);
-			vpninfo->urlpath = strdup(optarg);
+			free(urlpath);
+			urlpath = strdup(config_arg);
 			break;
 		case 'h':
 			usage();
 		case 'i':
-			vpninfo->ifname = optarg;
+			vpninfo->ifname = keep_config_arg();
 			break;
 		case 'l':
 			use_syslog = 1;
 			break;
 		case 'm':
-			vpninfo->mtu = atol(optarg);
+			vpninfo->mtu = atol(config_arg);
 			if (vpninfo->mtu < 576) {
 				fprintf(stderr, _("MTU %d too small\n"), vpninfo->mtu);
 				vpninfo->mtu = 576;
 			}
 			break;
 		case 'p':
-			vpninfo->cert_password = optarg;
+			vpninfo->cert_password = keep_config_arg();
 			break;
 		case 'P': 
-			proxy = optarg;
+			proxy = keep_config_arg();
 			autoproxy = 0;
 			break;
 		case OPT_NO_PROXY:
@@ -413,22 +572,22 @@ int main(int argc, char **argv)
 			nocertcheck = 1;
 			break;
 		case 's':
-			vpninfo->vpnc_script = optarg;
+			vpninfo->vpnc_script = keep_config_arg();
 			break;
 		case 'S':
 			vpninfo->script_tun = 1;
 			break;
 		case 'u':
-			vpninfo->username = optarg;
+			vpninfo->username = keep_config_arg();
 			break;
 		case 'U': {
 			char *strend;
-			uid = strtol(optarg, &strend, 0);
+			uid = strtol(config_arg, &strend, 0);
 			if (strend[0]) {
-				struct passwd *pw = getpwnam(optarg);
+				struct passwd *pw = getpwnam(config_arg);
 				if (!pw) {
 					fprintf(stderr, _("Invalid user \"%s\"\n"),
-						optarg);
+						config_arg);
 					exit(1);
 				}
 				uid = pw->pw_uid;
@@ -437,12 +596,12 @@ int main(int argc, char **argv)
 		}
 		case OPT_CSD_USER: {
 			char *strend;
-			vpninfo->uid_csd = strtol(optarg, &strend, 0);
+			vpninfo->uid_csd = strtol(config_arg, &strend, 0);
 			if (strend[0]) {
-				struct passwd *pw = getpwnam(optarg);
+				struct passwd *pw = getpwnam(config_arg);
 				if (!pw) {
 					fprintf(stderr, _("Invalid user \"%s\"\n"),
-						optarg);
+						config_arg);
 					exit(1);
 				}
 				vpninfo->uid_csd = pw->pw_uid;
@@ -451,13 +610,13 @@ int main(int argc, char **argv)
 			break;
 		}
 		case OPT_CSD_WRAPPER:
-			vpninfo->csd_wrapper = optarg;
+			vpninfo->csd_wrapper = keep_config_arg();
 			break;
 		case OPT_DISABLE_IPV6:
 			vpninfo->disable_ipv6 = 1;
 			break;
 		case 'Q':
-			vpninfo->max_qlen = atol(optarg);
+			vpninfo->max_qlen = atol(config_arg);
 			if (!vpninfo->max_qlen) {
 				fprintf(stderr, _("Queue length zero not permitted; using 1\n"));
 				vpninfo->max_qlen = 1;
@@ -470,10 +629,10 @@ int main(int argc, char **argv)
 			verbose = PRG_TRACE;
 			break;
 		case 'V':
-			printf(_("OpenConnect version %s\n"), openconnect_version);
+			printf(_("OpenConnect version %s\n"), openconnect_version_str);
 			exit(0);
 		case 'x':
-			vpninfo->xmlconfig = optarg;
+			vpninfo->xmlconfig = keep_config_arg();
 			vpninfo->write_new_config = write_new_config;
 			break;
 		case OPT_KEY_PASSWORD_FROM_FSID:
@@ -481,17 +640,20 @@ int main(int argc, char **argv)
 			break;
 		case OPT_USERAGENT:
 			free(vpninfo->useragent);
-			vpninfo->useragent = optarg;
+			vpninfo->useragent = strdup(config_arg);
 			break;
 		case OPT_FORCE_DPD:
-			vpninfo->dtls_times.dpd = vpninfo->ssl_times.dpd = atoi(optarg);
+			vpninfo->dtls_times.dpd = vpninfo->ssl_times.dpd = atoi(config_arg);
 			break;
 		default:
 			usage();
 		}
 	}
 
-	if (optind != argc - 1) {
+	if (optind < argc - 1) {
+		fprintf(stderr, _("Too many arguments on command line\n"));
+		usage();
+	} else if (optind > argc - 1) {
 		fprintf(stderr, _("No server specified\n"));
 		usage();
 	}
@@ -534,27 +696,22 @@ int main(int argc, char **argv)
 
 	if (!vpninfo->hostname) {
 		char *url = strdup(argv[optind]);
-		char *scheme;
-		char *group;
 
-		if (internal_parse_url(url, &scheme, &vpninfo->hostname, &vpninfo->port,
-			      &group, 443)) {
-			fprintf(stderr, _("Failed to parse server URL '%s'\n"),
-				url);
+		if (openconnect_parse_url(vpninfo, url))
 			exit(1);
-		}
-		if (scheme && strcmp(scheme, "https")) {
-			fprintf(stderr, _("Only https:// permitted for server URL\n"));
-			exit(1);
-		}
-		if (group) {
-			free(vpninfo->urlpath);
-			vpninfo->urlpath = group;
-		}
-		free(scheme);
+
 		free(url);
 	}
 
+	/* Historically, the path in the URL superseded the one in the
+	 * --usergroup argument, just because of the order in which they
+	 * were processed. Preserve that behaviour. */
+	if (urlpath && !vpninfo->urlpath) {
+		vpninfo->urlpath = urlpath;
+		urlpath = NULL;
+	}
+	free(urlpath);
+
 #ifdef SSL_UI
 	set_openssl_ui();
 #endif
@@ -582,7 +739,8 @@ int main(int argc, char **argv)
 
 	if (uid != getuid()) {
 		if (setuid(uid)) {
-			fprintf(stderr, _("Failed to set uid %d\n"), uid);
+			fprintf(stderr, _("Failed to set uid %ld\n"),
+				(long)uid);
 			exit(1);
 		}
 	}
@@ -599,9 +757,12 @@ int main(int argc, char **argv)
 		     (vpninfo->deflate ? "SSL + deflate" : "SSL")
 		     : "DTLS");
 
-	if (!vpninfo->vpnc_script)
+	if (!vpninfo->vpnc_script) {
 		vpn_progress(vpninfo, PRG_INFO,
 			     _("No --script argument provided; DNS and routing are not configured\n"));
+		vpn_progress(vpninfo, PRG_INFO,
+			     _("See http://www.infradead.org/openconnect/vpnc-script.html\n"));
+	}
 
 	if (background) {
 		int pid;
@@ -671,6 +832,7 @@ void write_progress(void *_vpninfo, int level, const char *fmt, ...)
 		va_start(args, fmt);
 		vfprintf(outf, fmt, args);
 		va_end(args);
+		fflush(outf);
 	}
 }
 
@@ -745,12 +907,16 @@ static int validate_peer_cert(void *_vpninfo, X509 *peer_cert,
 		fprintf(stderr,
 			_("\nCertificate from VPN server \"%s\" failed verification.\n"
 			  "Reason: %s\n"), vpninfo->hostname, reason);
+		fflush(stderr);
+
 		if (non_inter)
 			return -EINVAL;
 
-		fflush(stderr);
-
 		ui = UI_new();
+		if (!ui) {
+			fprintf(stderr, _("Failed to create UI\n"));
+			return -EINVAL;
+		}
 		UI_add_input_string(ui, _("Enter 'yes' to accept, 'no' to abort; anything else to view: "),
 				    UI_INPUT_FLAG_ECHO, buf, 2, 5);
 		ret = UI_process(ui);
