@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2011 Intel Corporation.
+ * Copyright © 2008-2012 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
@@ -135,6 +135,8 @@ struct openconnect_info {
 	int uid_csd_given;
 	int no_http_keepalive;
 
+	X509 *peer_cert;
+
 	char *cookie; /* Pointer to within cookies list */
 	struct vpn_option *cookies;
 	struct vpn_option *cstp_options;
@@ -146,6 +148,7 @@ struct openconnect_info {
 	int owe_ssl_dpd_response;
 	struct pkt *deflate_pkt;
 	struct pkt *current_ssl_pkt;
+	struct pkt *pending_deflated_pkt;
 
 	z_stream inflate_strm;
 	uint32_t inflate_adler32;
@@ -166,7 +169,7 @@ struct openconnect_info {
 	unsigned char dtls_secret[48];
 
 	char *dtls_cipher;
-	char *vpnc_script;
+	const char *vpnc_script;
 	int script_tun;
 	char *ifname;
 
@@ -190,12 +193,13 @@ struct openconnect_info {
 
 #ifdef __sun__
 	int ip_fd;
-	int tun_muxid;
+	int ip6_fd;
 #endif
 	int tun_fd;
 	int ssl_fd;
 	int dtls_fd;
 	int new_dtls_fd;
+	int cancel_fd;
 
 	struct pkt *incoming_queue;
 	struct pkt *outgoing_queue;
@@ -237,13 +241,28 @@ struct openconnect_info {
 
 #define vpn_progress(vpninfo, ...) (vpninfo)->progress ((vpninfo)->cbdata, __VA_ARGS__)
 
+/****************************************************************************/
+/* Oh Solaris how we hate thee! */
+#ifdef __sun__
+#define time(x) openconnect__time(x)
+time_t openconnect__time(time_t *t);
+#endif
+#ifndef HAVE_ASPRINTF
+#define asprintf openconnect__asprintf
+int openconnect__asprintf(char **strp, const char *fmt, ...);
+#endif
+#ifndef HAVE_GETLINE
+#define getline openconnect__getline
+ssize_t openconnect__getline(char **lineptr, size_t *n, FILE *stream);
+#endif
+
 /****************************************************************************/
 
 /* tun.c */
 int setup_tun(struct openconnect_info *vpninfo);
 int tun_mainloop(struct openconnect_info *vpninfo, int *timeout);
 void shutdown_tun(struct openconnect_info *vpninfo);
-void script_reconnect (struct openconnect_info *vpninfo);
+int script_config_tun (struct openconnect_info *vpninfo, const char *reason);
 
 /* dtls.c */
 unsigned char unhex(const char *data);
@@ -260,13 +279,15 @@ int cstp_reconnect(struct openconnect_info *vpninfo);
 
 /* ssl.c */
 int  __attribute__ ((format (printf, 2, 3)))
-		openconnect_SSL_printf(SSL *ssl, const char *fmt, ...);
-int openconnect_SSL_gets(SSL *ssl, char *buf, size_t len);
+    openconnect_SSL_printf(struct openconnect_info *vpninfo, const char *fmt, ...);
+int openconnect_SSL_gets(struct openconnect_info *vpninfo, char *buf, size_t len);
+int openconnect_SSL_write(struct openconnect_info *vpninfo, char *buf, size_t len);
+int openconnect_SSL_read(struct openconnect_info *vpninfo, char *buf, size_t len);
 int openconnect_open_https(struct openconnect_info *vpninfo);
 void openconnect_close_https(struct openconnect_info *vpninfo);
 int get_cert_md5_fingerprint(struct openconnect_info *vpninfo, X509 *cert,
 			     char *buf);
-void report_ssl_errors(struct openconnect_info *vpninfo);
+void openconnect_report_ssl_errors(struct openconnect_info *vpninfo);
 
 /* mainloop.c */
 int vpn_add_pollfd(struct openconnect_info *vpninfo, int fd, short events);
@@ -300,6 +321,6 @@ int generate_securid_tokencodes(struct openconnect_info *vpninfo);
 int add_securid_pin(char *token, char *pin);
 
 /* version.c */
-extern char openconnect_version[];
+extern const char *openconnect_version_str;
 
 #endif /* __OPENCONNECT_INTERNAL_H__ */
