@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2010 Intel Corporation.
+ * Copyright © 2008-2012 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -60,8 +60,86 @@
 /* Helper functions for reading/writing lines over SSL.
    We could use cURL for the HTTP stuff, but it's overkill */
 
+int openconnect_SSL_write(struct openconnect_info *vpninfo, char *buf, size_t len)
+{
+	size_t orig_len = len;
+
+	while (len) {
+		int done = SSL_write(vpninfo->https_ssl, buf, len);
+
+		if (done > 0)
+			len -= done;
+		else {
+			int err = SSL_get_error(vpninfo->https_ssl, done);
+			fd_set wr_set, rd_set;
+			int maxfd = vpninfo->ssl_fd;
+
+			FD_ZERO(&wr_set);
+			FD_ZERO(&rd_set);
+			
+			if (err == SSL_ERROR_WANT_READ)
+				FD_SET(vpninfo->ssl_fd, &rd_set);
+			else if (err == SSL_ERROR_WANT_WRITE)
+				FD_SET(vpninfo->ssl_fd, &wr_set);
+			else {
+				vpn_progress(vpninfo, PRG_ERR, _("Failed to write to SSL socket"));
+				openconnect_report_ssl_errors(vpninfo);
+				return -EIO;
+			}
+			if (vpninfo->cancel_fd != -1) {
+				FD_SET(vpninfo->cancel_fd, &rd_set);
+				if (vpninfo->cancel_fd > vpninfo->ssl_fd)
+					maxfd = vpninfo->cancel_fd;
+			}
+			select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+			if (vpninfo->cancel_fd != -1 &&
+			    FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+				vpn_progress(vpninfo, PRG_ERR, _("SSL write cancelled\n"));
+				return -EINTR;
+			}
+		}
+	}
+	return orig_len;
+}
+
+int openconnect_SSL_read(struct openconnect_info *vpninfo, char *buf, size_t len)
+{
+	int done;
+
+	while ((done = SSL_read(vpninfo->https_ssl, buf, len)) == -1) {
+		int err = SSL_get_error(vpninfo->https_ssl, done);
+		fd_set wr_set, rd_set;
+		int maxfd = vpninfo->ssl_fd;
+
+		FD_ZERO(&wr_set);
+		FD_ZERO(&rd_set);
+			
+		if (err == SSL_ERROR_WANT_READ)
+			FD_SET(vpninfo->ssl_fd, &rd_set);
+		else if (err == SSL_ERROR_WANT_WRITE)
+			FD_SET(vpninfo->ssl_fd, &wr_set);
+		else {
+			vpn_progress(vpninfo, PRG_ERR, _("Failed to read from SSL socket"));
+			openconnect_report_ssl_errors(vpninfo);
+			return -EIO;
+		}
+		if (vpninfo->cancel_fd != -1) {
+			FD_SET(vpninfo->cancel_fd, &rd_set);
+			if (vpninfo->cancel_fd > vpninfo->ssl_fd)
+				maxfd = vpninfo->cancel_fd;
+		}
+		select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+		if (vpninfo->cancel_fd != -1 &&
+		    FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+			vpn_progress(vpninfo, PRG_ERR, _("SSL read cancelled\n"));
+			return -EINTR;
+		}
+	}
+	return done;
+}
+
 int  __attribute__ ((format (printf, 2, 3)))
-	openconnect_SSL_printf(SSL *ssl, const char *fmt, ...)
+    openconnect_SSL_printf(struct openconnect_info *vpninfo, const char *fmt, ...)
 {
 	char buf[1024];
 	va_list args;
@@ -71,7 +149,7 @@ int  __attribute__ ((format (printf, 2, 3)))
 	va_start(args, fmt);
 	vsnprintf(buf, 1023, fmt, args);
 	va_end(args);
-	return SSL_write(ssl, buf, strlen(buf));
+	return openconnect_SSL_write(vpninfo, buf, strlen(buf));
 
 }
 
@@ -83,12 +161,12 @@ static int print_err(const char *str, size_t len, void *ptr)
 	return 0;
 }
 
-void report_ssl_errors(struct openconnect_info *vpninfo)
+void openconnect_report_ssl_errors(struct openconnect_info *vpninfo)
 {
 	ERR_print_errors_cb(print_err, vpninfo);
 }
 
-int openconnect_SSL_gets(SSL *ssl, char *buf, size_t len)
+int openconnect_SSL_gets(struct openconnect_info *vpninfo, char *buf, size_t len)
 {
 	int i = 0;
 	int ret;
@@ -96,25 +174,55 @@ int openconnect_SSL_gets(SSL *ssl, char *buf, size_t len)
 	if (len < 2)
 		return -EINVAL;
 
-	while ( (ret = SSL_read(ssl, buf + i, 1)) == 1) {
-		if (buf[i] == '\n') {
-			buf[i] = 0;
-			if (i && buf[i-1] == '\r') {
-				buf[i-1] = 0;
-				i--;
+	while (1) {
+		ret = SSL_read(vpninfo->https_ssl, buf + i, 1);
+		if (ret == 1) {
+			if (buf[i] == '\n') {
+				buf[i] = 0;
+				if (i && buf[i-1] == '\r') {
+					buf[i-1] = 0;
+					i--;
+				}
+				return i;
 			}
-			return i;
-		}
-		i++;
+			i++;
 
-		if (i >= len - 1) {
-			buf[i] = 0;
-			return i;
+			if (i >= len - 1) {
+				buf[i] = 0;
+				return i;
+			}
+		} else {
+			fd_set rd_set, wr_set;
+			int maxfd = vpninfo->ssl_fd;
+			
+			FD_ZERO(&rd_set);
+			FD_ZERO(&wr_set);
+			
+			ret = SSL_get_error(vpninfo->https_ssl, ret);
+			if (ret == SSL_ERROR_WANT_READ)
+				FD_SET(vpninfo->ssl_fd, &rd_set);
+			else if (ret == SSL_ERROR_WANT_WRITE)
+				FD_SET(vpninfo->ssl_fd, &wr_set);
+			else {
+				vpn_progress(vpninfo, PRG_ERR, _("Failed to read from SSL socket\n"));
+				openconnect_report_ssl_errors(vpninfo);
+				ret = -EIO;
+				break;
+			}
+			if (vpninfo->cancel_fd != -1) {
+				FD_SET(vpninfo->cancel_fd, &rd_set);
+				if (vpninfo->cancel_fd > vpninfo->ssl_fd)
+					maxfd = vpninfo->cancel_fd;
+			}
+			select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+			if (vpninfo->cancel_fd != -1 &&
+			    FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+				vpn_progress(vpninfo, PRG_ERR, _("SSL read cancelled\n"));
+				ret = -EINTR;
+				break;
+			}
 		}
 	}
-	if (ret == 0) {
-		ret = -SSL_get_error(ssl, ret);
-	}
 	buf[i] = 0;
 	return i ?: ret;
 }
@@ -159,7 +267,7 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 	if (!PKCS12_parse(p12, vpninfo->cert_password?:pass, &pkey, &cert, &ca)) {
 		unsigned long err = ERR_peek_error();
 
-		report_ssl_errors(vpninfo);
+		openconnect_report_ssl_errors(vpninfo);
 
 		if (ERR_GET_LIB(err) == ERR_LIB_PKCS12 &&
 		    ERR_GET_FUNC(err) == PKCS12_F_PKCS12_PARSE &&
@@ -232,13 +340,13 @@ static int load_tpm_certificate(struct openconnect_info *vpninfo)
 	e = ENGINE_by_id("tpm");
 	if (!e) {
 		vpn_progress(vpninfo, PRG_ERR, _("Can't load TPM engine.\n"));
-		report_ssl_errors(vpninfo);
+		openconnect_report_ssl_errors(vpninfo);
 		return -EINVAL;
 	}
 	if (!ENGINE_init(e) || !ENGINE_set_default_RSA(e) ||
 	    !ENGINE_set_default_RAND(e)) {
 		vpn_progress(vpninfo, PRG_ERR, _("Failed to init TPM engine\n"));
-		report_ssl_errors(vpninfo);
+		openconnect_report_ssl_errors(vpninfo);
 		ENGINE_free(e);
 		return -EINVAL;
 	}
@@ -248,21 +356,21 @@ static int load_tpm_certificate(struct openconnect_info *vpninfo)
 				     vpninfo->cert_password, NULL, 0)) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Failed to set TPM SRK password\n"));
-			report_ssl_errors(vpninfo);
+			openconnect_report_ssl_errors(vpninfo);
 		}
 	}
 	key = ENGINE_load_private_key(e, vpninfo->sslkey, NULL, NULL);
 	if (!key) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to load TPM private key\n"));
-		report_ssl_errors(vpninfo);
+		openconnect_report_ssl_errors(vpninfo);
 		ENGINE_free(e);
 		ENGINE_finish(e);
 		return -EINVAL;
 	}
 	if (!SSL_CTX_use_PrivateKey(vpninfo->https_ctx, key)) {
 		vpn_progress(vpninfo, PRG_ERR, _("Add key from TPM failed\n"));
-		report_ssl_errors(vpninfo);
+		openconnect_report_ssl_errors(vpninfo);
 		ENGINE_free(e);
 		ENGINE_finish(e);
 		return -EINVAL;
@@ -290,7 +398,7 @@ static int reload_pem_cert(struct openconnect_info *vpninfo)
 		BIO_free(b);
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to reload X509 cert for expiry check\n"));
-		report_ssl_errors(vpninfo);
+		openconnect_report_ssl_errors(vpninfo);
 		return -EIO;
 	}
 	vpninfo->cert_x509 = PEM_read_bio_X509_AUX(b, NULL, NULL, NULL);
@@ -325,7 +433,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		/* Not PKCS#12 */
 		if (vpninfo->cert_type == CERT_TYPE_PKCS12) {
 			vpn_progress(vpninfo, PRG_ERR, _("Read PKCS#12 failed\n"));
-			report_ssl_errors(vpninfo);
+			openconnect_report_ssl_errors(vpninfo);
 			return -EINVAL;
 		}
 		/* Clear error and fall through to see if it's a PEM file... */
@@ -337,7 +445,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 						vpninfo->cert)) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Loading certificate failed\n"));
-		report_ssl_errors(vpninfo);
+		openconnect_report_ssl_errors(vpninfo);
 		return -EINVAL;
 	}
 
@@ -391,7 +499,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 					    SSL_FILETYPE_PEM)) {
 		unsigned long err = ERR_peek_error();
 		
-		report_ssl_errors(vpninfo);
+		openconnect_report_ssl_errors(vpninfo);
 
 #ifndef EVP_F_EVP_DECRYPTFINAL_EX
 #define EVP_F_EVP_DECRYPTFINAL_EX EVP_F_EVP_DECRYPTFINAL
@@ -862,6 +970,42 @@ static int check_certificate_expiry(struct openconnect_info *vpninfo)
 	return 0;
 }
 
+static int cancellable_connect(struct openconnect_info *vpninfo, int sockfd,
+			       const struct sockaddr *addr, socklen_t addrlen)
+{
+	struct sockaddr_storage peer;
+	socklen_t peerlen = sizeof(peer);
+	fd_set wr_set, rd_set;
+	int maxfd = sockfd;
+
+	fcntl(sockfd, F_SETFL, fcntl(sockfd, F_GETFL) | O_NONBLOCK);
+
+	if (connect(sockfd, addr, addrlen) < 0 && errno != EINPROGRESS)
+		return -1;
+
+	FD_ZERO(&wr_set);
+	FD_ZERO(&rd_set);
+	FD_SET(sockfd, &wr_set);
+	if (vpninfo->cancel_fd != -1) {
+		FD_SET(vpninfo->cancel_fd, &rd_set);
+		if (vpninfo->cancel_fd > sockfd)
+			maxfd = vpninfo->cancel_fd;
+	}
+	
+	/* Later we'll render this whole exercise non-pointless by
+	   including a 'cancelfd' here too. */
+	select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+	if (vpninfo->cancel_fd != -1 && FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+		vpn_progress(vpninfo, PRG_ERR, _("Socket connect cancelled\n"));
+		errno = EINTR;
+		return -1;
+	}
+		
+	/* Check whether connect() succeeded or failed by using
+	   getpeername(). See http://cr.yp.to/docs/connect.html */
+	return getpeername(sockfd, (void *)&peer, &peerlen);
+}
+
 int openconnect_open_https(struct openconnect_info *vpninfo)
 {
 	method_const SSL_METHOD *ssl3_method;
@@ -874,8 +1018,17 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		vpninfo->port = 443;
 
 	if (vpninfo->peer_addr) {
-		ssl_sock = socket(vpninfo->peer_addr->sa_family, SOCK_STREAM, IPPROTO_IP);
-		if (ssl_sock < 0) {
+#ifdef SOCK_CLOEXEC
+		ssl_sock = socket(vpninfo->peer_addr->sa_family, SOCK_STREAM | SOCK_CLOEXEC, IPPROTO_IP);
+		if (ssl_sock < 0)
+#endif
+		{
+			ssl_sock = socket(vpninfo->peer_addr->sa_family, SOCK_STREAM, IPPROTO_IP);
+			if (ssl_sock < 0)
+				goto reconn_err;
+			fcntl(ssl_sock, F_SETFD, fcntl(ssl_sock, F_GETFD) | FD_CLOEXEC);
+		}
+		if (cancellable_connect(vpninfo, ssl_sock, vpninfo->peer_addr, vpninfo->peer_addrlen)) {
 		reconn_err:
 			if (vpninfo->proxy) {
 				vpn_progress(vpninfo, PRG_ERR, 
@@ -888,8 +1041,6 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			}
 			return -EINVAL;
 		}
-		if (connect(ssl_sock, vpninfo->peer_addr, vpninfo->peer_addrlen))
-			goto reconn_err;
 		
 	} else {
 		struct addrinfo hints, *result, *rp;
@@ -932,6 +1083,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			proxies = px_proxy_factory_get_proxies(vpninfo->proxy_factory,
 							       url);
 
+			i = 0;
 			while (proxies && proxies[i]) {
 				if (!vpninfo->proxy &&
 				    (!strncmp(proxies[i], "http://", 7) ||
@@ -998,7 +1150,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 					  rp->ai_protocol);
 			if (ssl_sock < 0)
 				continue;
-			if (connect(ssl_sock, rp->ai_addr, rp->ai_addrlen) >= 0) {
+			if (cancellable_connect(vpninfo, ssl_sock, rp->ai_addr, rp->ai_addrlen) >= 0) {
 				/* Store the peer address we actually used, so that DTLS can
 				   use it again later */
 				vpninfo->peer_addr = malloc(rp->ai_addrlen);
@@ -1024,7 +1176,6 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			return -EINVAL;
 		}
 	}
-	fcntl(ssl_sock, F_SETFD, FD_CLOEXEC);
 
 	if (vpninfo->proxy) {
 		err = process_proxy(vpninfo, ssl_sock);
@@ -1073,7 +1224,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Failed to open CA file '%s'\n"),
 					     vpninfo->cafile);
-				report_ssl_errors(vpninfo);
+				openconnect_report_ssl_errors(vpninfo);
 				close(ssl_sock);
 				return -EINVAL;
 			}
@@ -1084,17 +1235,45 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	workaround_openssl_certchain_bug(vpninfo, https_ssl);
 
 	https_bio = BIO_new_socket(ssl_sock, BIO_NOCLOSE);
+	BIO_set_nbio(https_bio, 1);
 	SSL_set_bio(https_ssl, https_bio, https_bio);
 
 	vpn_progress(vpninfo, PRG_INFO, _("SSL negotiation with %s\n"),
 		     vpninfo->hostname);
 
-	if (SSL_connect(https_ssl) <= 0) {
-		vpn_progress(vpninfo, PRG_ERR, _("SSL connection failure\n"));
-		report_ssl_errors(vpninfo);
-		SSL_free(https_ssl);
-		close(ssl_sock);
-		return -EINVAL;
+	while ((err = SSL_connect(https_ssl)) <= 0) {
+		fd_set wr_set, rd_set;
+		int maxfd = ssl_sock;
+		
+		FD_ZERO(&wr_set);
+		FD_ZERO(&rd_set);
+
+		err = SSL_get_error(https_ssl, err);
+		if (err == SSL_ERROR_WANT_READ)
+			FD_SET(ssl_sock, &rd_set);
+		else if (err == SSL_ERROR_WANT_WRITE)
+			FD_SET(ssl_sock, &wr_set);
+		else {
+			vpn_progress(vpninfo, PRG_ERR, _("SSL connection failure\n"));
+			openconnect_report_ssl_errors(vpninfo);
+			SSL_free(https_ssl);
+			close(ssl_sock);
+			return -EINVAL;
+		}
+
+		if (vpninfo->cancel_fd != -1) {
+			FD_SET(vpninfo->cancel_fd, &rd_set);
+			if (vpninfo->cancel_fd > ssl_sock)
+				maxfd = vpninfo->cancel_fd;
+		}
+		select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+		if (vpninfo->cancel_fd != -1 &&
+		    FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+			vpn_progress(vpninfo, PRG_ERR, _("SSL connection cancelled\n"));
+			SSL_free(https_ssl);
+			close(ssl_sock);
+			return -EINVAL;
+		}
 	}
 
 	if (verify_peer(vpninfo, https_ssl)) {
@@ -1114,6 +1293,10 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 
 void openconnect_close_https(struct openconnect_info *vpninfo)
 {
+	if (vpninfo->peer_cert) {
+		X509_free(vpninfo->peer_cert);
+		vpninfo->peer_cert = NULL;
+	}
 	SSL_free(vpninfo->https_ssl);
 	vpninfo->https_ssl = NULL;
 	close(vpninfo->ssl_fd);
