@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2010 Intel Corporation.
+ * Copyright © 2008-2012 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -75,17 +75,9 @@
 #define TUN_HAS_AF_PREFIX 1
 #endif
 
-#ifdef __sun__
-static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
-{
-	if (!mtu_only)
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("No vpnc-script configured. Need Solaris IP-setting code\n"));
-	return 0;
-}
-#else
-static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
+static int set_tun_mtu(struct openconnect_info *vpninfo)
 {
+#ifndef __sun__ /* We don't know how to do this on Solaris */
 	struct ifreq ifr;
 	int net_fd;
 
@@ -94,35 +86,19 @@ static int local_config_tun(struct openconnect_info *vpninfo, int mtu_only)
 		perror(_("open net"));
 		return -EINVAL;
 	}
+
 	memset(&ifr, 0, sizeof(ifr));
 	strncpy(ifr.ifr_name, vpninfo->ifname, sizeof(ifr.ifr_name) - 1);
-
-	if (!mtu_only) {
-		struct sockaddr_in addr;
-
-		if (ioctl(net_fd, SIOCGIFFLAGS, &ifr) < 0)
-			perror(_("SIOCGIFFLAGS"));
-
-		ifr.ifr_flags |= IFF_UP | IFF_POINTOPOINT;
-		if (ioctl(net_fd, SIOCSIFFLAGS, &ifr) < 0)
-			perror(_("SIOCSIFFLAGS"));
-
-		addr.sin_family = AF_INET;
-		addr.sin_addr.s_addr = inet_addr(vpninfo->vpn_addr);
-		memcpy(&ifr.ifr_addr, &addr, sizeof(addr));
-		if (ioctl(net_fd, SIOCSIFADDR, &ifr) < 0)
-			perror(_("SIOCSIFADDR"));
-	}
-
 	ifr.ifr_mtu = vpninfo->mtu;
+
 	if (ioctl(net_fd, SIOCSIFMTU, &ifr) < 0)
 		perror(_("SIOCSIFMTU"));
 
 	close(net_fd);
-
+#endif
 	return 0;
 }
-#endif
+
 
 static int setenv_int(const char *opt, int value)
 {
@@ -248,7 +224,7 @@ static void set_banner(struct openconnect_info *vpninfo)
 	char *banner, *q;
 	const char *p;
 
-	if (!vpninfo->banner || !(banner = malloc(strlen(vpninfo->banner)))) {
+	if (!vpninfo->banner || !(banner = malloc(strlen(vpninfo->banner)+1))) {
 		unsetenv("CISCO_BANNER");
 		return;
 	}
@@ -277,7 +253,6 @@ static void set_script_env(struct openconnect_info *vpninfo)
 	if (!ret)
 		setenv("VPNGATEWAY", host, 1);
 
-	setenv("reason", "connect", 1);
 	set_banner(vpninfo);
 	unsetenv("CISCO_SPLIT_INC");
 	unsetenv("CISCO_SPLIT_EXC");
@@ -368,22 +343,179 @@ static void set_script_env(struct openconnect_info *vpninfo)
 	setenv_cstp_opts(vpninfo);
 }
 
-static int script_config_tun(struct openconnect_info *vpninfo)
+int script_config_tun(struct openconnect_info *vpninfo, const char *reason)
 {
+	if (!vpninfo->vpnc_script)
+		return 0;
+
+	setenv("reason", reason, 1);
 	if (system(vpninfo->vpnc_script)) {
 		int e = errno;
 		vpn_progress(vpninfo, PRG_ERR,
-			     _("Failed to spawn script '%s': %s\n"),
-			     vpninfo->vpnc_script, strerror(e));
+			     _("Failed to spawn script '%s' for %s: %s\n"),
+			     vpninfo->vpnc_script, reason, strerror(e));
 		return -e;
 	}
 	return 0;
 }
 
-void script_reconnect (struct openconnect_info *vpninfo)
+#ifdef __sun__
+static int link_proto(int unit_nr, const char *devname, uint64_t flags)
 {
-	setenv("reason", "reconnect", 1);
-	script_config_tun(vpninfo);
+	int ip_fd, mux_id, tun2_fd;
+	struct lifreq ifr;
+
+	tun2_fd = open("/dev/tun", O_RDWR);
+	if (tun2_fd < 0) {
+		perror(_("Could not /dev/tun for plumbing"));
+		return -EIO;
+	}
+	if (ioctl(tun2_fd, I_PUSH, "ip") < 0) {
+		perror(_("Can't push IP"));
+		close(tun2_fd);
+		return -EIO;
+	}
+
+	sprintf(ifr.lifr_name, "tun%d", unit_nr);
+	ifr.lifr_ppa = unit_nr;
+	ifr.lifr_flags = flags;
+
+	if (ioctl(tun2_fd, SIOCSLIFNAME, &ifr) < 0) {
+		perror(_("Can't set ifname"));
+		close(tun2_fd);
+		return -1;
+	}
+
+	ip_fd = open(devname, O_RDWR);
+	if (ip_fd < 0) {
+		fprintf(stderr, _("Can't open %s: %s"), devname,
+			strerror(errno));
+		close(tun2_fd);
+		return -1;
+	}
+
+	mux_id = ioctl(ip_fd, I_LINK, tun2_fd);
+	if (mux_id < 0) {
+		fprintf(stderr, _("Can't plumb %s for IPv%d: %s\n"),
+			 ifr.lifr_name, (flags == IFF_IPV4) ? 4 : 6,
+			 strerror(errno));
+		close(tun2_fd);
+		close(ip_fd);
+		return -1;
+	}
+
+	close(tun2_fd);
+
+	return ip_fd;
+}
+#endif
+
+static int os_setup_tun(struct openconnect_info *vpninfo)
+{
+	int tun_fd;
+
+#ifdef IFF_TUN /* Linux */
+	struct ifreq ifr;
+	int tunerr;
+
+	tun_fd = open("/dev/net/tun", O_RDWR);
+	if (tun_fd < 0) {
+		/* Android has /dev/tun instead of /dev/net/tun
+		   Since other systems might have too, just try it
+		   as a fallback instead of using ifdef __ANDROID__ */
+		tunerr = errno;
+		tun_fd = open("/dev/tun", O_RDWR);
+	}
+	if (tun_fd < 0) {
+		/* If the error on /dev/tun is ENOENT, that's boring.
+		   Use the error we got on /dev/net/tun instead */
+		if (errno != -ENOENT)
+			tunerr = errno;
+
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to open tun device: %s\n"),
+			     strerror(tunerr));
+		exit(1);
+	}
+	memset(&ifr, 0, sizeof(ifr));
+	ifr.ifr_flags = IFF_TUN | IFF_NO_PI;
+	if (vpninfo->ifname)
+		strncpy(ifr.ifr_name, vpninfo->ifname,
+			sizeof(ifr.ifr_name) - 1);
+	if (ioctl(tun_fd, TUNSETIFF, (void *) &ifr) < 0) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("TUNSETIFF failed: %s\n"),
+			     strerror(errno));
+		exit(1);
+	}
+	if (!vpninfo->ifname)
+		vpninfo->ifname = strdup(ifr.ifr_name);
+#elif defined (__sun__)
+	static char tun_name[80];
+	int unit_nr;
+
+	tun_fd = open("/dev/tun", O_RDWR);
+	if (tun_fd < 0) {
+		perror(_("open /dev/tun"));
+		return -EIO;
+	}
+
+	unit_nr = ioctl(tun_fd, TUNNEWPPA, -1);
+	if (unit_nr < 0) {
+		perror(_("Failed to create new tun"));
+		close(tun_fd);
+		return -EIO;
+	}
+
+	if (ioctl(tun_fd, I_SRDOPT, RMSGD) < 0) {
+		perror(_("Failed to put tun file descriptor into message-discard mode"));
+		close(tun_fd);
+		return -EIO;
+	}
+
+	sprintf(tun_name, "tun%d", unit_nr);
+	vpninfo->ifname = strdup(tun_name);
+
+	vpninfo->ip_fd = link_proto(unit_nr, "/dev/udp", IFF_IPV4);
+	if (vpninfo->ip_fd < 0) {
+		close(tun_fd);
+		return -EIO;
+	}
+
+	if (vpninfo->vpn_addr6) {
+		vpninfo->ip6_fd = link_proto(unit_nr, "/dev/udp6", IFF_IPV6);
+		if (vpninfo->ip6_fd < 0) {
+			close(tun_fd);
+			close(vpninfo->ip_fd);
+			vpninfo->ip_fd = -1;
+			return -EIO;
+		}
+	} else
+		vpninfo->ip6_fd = -1;
+
+#else /* BSD et al have /dev/tun$x devices */
+	static char tun_name[80];
+	int i;
+	for (i = 0; i < 255; i++) {
+		sprintf(tun_name, "/dev/tun%d", i);
+		tun_fd = open(tun_name, O_RDWR);
+		if (tun_fd >= 0)
+			break;
+	}
+	if (tun_fd < 0) {
+		perror(_("open tun"));
+		exit(1);
+	}
+	vpninfo->ifname = strdup(tun_name + 5);
+#ifdef TUNSIFHEAD
+	i = 1;
+	if (ioctl(tun_fd, TUNSIFHEAD, &i) < 0) {
+		perror(_("TUNSIFHEAD"));
+		exit(1);
+	}
+#endif
+#endif
+	return tun_fd;
 }
 
 /* Set up a tuntap device. */
@@ -417,147 +549,17 @@ int setup_tun(struct openconnect_info *vpninfo)
 		vpninfo->script_tun = child;
 		vpninfo->ifname = strdup(_("(script)"));
 	} else {
-#ifdef IFF_TUN /* Linux */
-		struct ifreq ifr;
-		int tunerr;
+		script_config_tun(vpninfo, "pre-init");
 
-		tun_fd = open("/dev/net/tun", O_RDWR);
-		if (tun_fd < 0) {
-			/* Android has /dev/tun instead of /dev/net/tun
-			   Since other systems might have too, just try it
-			   as a fallback instead of using ifdef __ANDROID__ */
-			tunerr = errno;
-			tun_fd = open("/dev/tun", O_RDWR);
-		}
-		if (tun_fd < 0) {
-			/* If the error on /dev/tun is ENOENT, that's boring.
-			   Use the error we got on /dev/net/tun instead */
-			if (errno != -ENOENT)
-				tunerr = errno;
+		tun_fd = os_setup_tun(vpninfo);
+		if (tun_fd < 0)
+			return tun_fd;
 
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Failed to open tun device: %s\n"),
-				     strerror(tunerr));
-			exit(1);
-		}
-		memset(&ifr, 0, sizeof(ifr));
-		ifr.ifr_flags = IFF_TUN | IFF_NO_PI;
-		if (vpninfo->ifname)
-			strncpy(ifr.ifr_name, vpninfo->ifname,
-				sizeof(ifr.ifr_name) - 1);
-		if (ioctl(tun_fd, TUNSETIFF, (void *) &ifr) < 0) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("TUNSETIFF failed: %s\n"),
-				     strerror(errno));
-			exit(1);
-		}
-		if (!vpninfo->ifname)
-			vpninfo->ifname = strdup(ifr.ifr_name);
-#elif defined (__sun__)
-		static char tun_name[80];
-		int tun2_fd, ip_fd = open("/dev/ip", O_RDWR);
-		int unit_nr, mux_id;
-		struct ifreq ifr;
+		setenv("TUNDEV", vpninfo->ifname, 1);
+		script_config_tun(vpninfo, "connect");
 
-		if (ip_fd < 0) {
-			perror(_("open /dev/ip"));
-			return -EIO;
-		}
-
-		tun_fd = open("/dev/tun", O_RDWR);
-		if (tun_fd < 0) {
-			perror(_("open /dev/tun"));
-			close(ip_fd);
-			return -EIO;
-		}
-
-		unit_nr = ioctl(tun_fd, TUNNEWPPA, -1);
-		if (unit_nr < 0) {
-			perror(_("Failed to create new tun"));
-			close(tun_fd);
-			close(ip_fd);
-			return -EIO;
-		}
-		
-		tun2_fd = open("/dev/tun", O_RDWR);
-		if (tun2_fd < 0) {
-			perror(_("open /dev/tun again"));
-			close(tun_fd);
-			close(ip_fd);
-			return -EIO;
-		}
-		if (ioctl(tun2_fd, I_PUSH, "ip") < 0) {
-			perror(_("Can't push IP"));
-			close(tun2_fd);
-			close(tun_fd);
-			close(ip_fd);
-			return -EIO;
-		}
-		if (ioctl(tun2_fd, IF_UNITSEL, &unit_nr) < 0) {
-			perror(_("Can't select unit"));
-			close(tun2_fd);
-			close(tun_fd);
-			close(ip_fd);
-			return -EIO;
-		}
-		mux_id = ioctl(ip_fd, I_PLINK, tun2_fd);
-		if (mux_id < 0) {
-			perror(_("Can't link tun to IP"));
-			close(tun2_fd);
-			close(tun_fd);
-			close(ip_fd);
-			return -EIO;
-		}
-		close(tun2_fd);
-
-		sprintf(tun_name, "tun%d", unit_nr);
-		vpninfo->ifname = strdup(tun_name);
-
-		memset(&ifr, 0, sizeof(ifr));
-		strcpy(ifr.ifr_name, tun_name);
-		ifr.ifr_ip_muxid = mux_id;
-
-		if (ioctl(ip_fd, SIOCSIFMUXID, &ifr) < 0) {
-			perror(_("Set mux id"));
-			close(tun_fd);
-			ioctl(ip_fd, I_PUNLINK, mux_id);
-			close(ip_fd);
-			return -EIO;
-		}
-		/* Solaris tunctl needs this in order to tear it down */
-		vpn_progress(vpninfo, PRG_DEBUG, _("mux id is %d\n"), mux_id);
-		vpninfo->tun_muxid = mux_id;
-		vpninfo->ip_fd = ip_fd;
-
-#else /* BSD et al have /dev/tun$x devices */
-		static char tun_name[80];
-		int i;
-		for (i = 0; i < 255; i++) {
-			sprintf(tun_name, "/dev/tun%d", i);
-			tun_fd = open(tun_name, O_RDWR);
-			if (tun_fd >= 0)
-				break;
-		}
-		if (tun_fd < 0) {
-			perror(_("open tun"));
-			exit(1);
-		}
-		vpninfo->ifname = strdup(tun_name + 5);
-#ifdef TUNSIFHEAD
-		i = 1;
-		if (ioctl(tun_fd, TUNSIFHEAD, &i) < 0) {
-			perror(_("TUNSIFHEAD"));
-			exit(1);
-		}
-#endif
-#endif
-		if (vpninfo->vpnc_script) {
-			setenv("TUNDEV", vpninfo->ifname, 1);
-			script_config_tun(vpninfo);
-			/* We have to set the MTU for ourselves, because the script doesn't */
-			local_config_tun(vpninfo, 1);
-		} else
-			local_config_tun(vpninfo, 0);
+		/* Ancient vpnc-scripts might not get this right */
+		set_tun_mtu(vpninfo);
 	}
 
 	fcntl(tun_fd, F_SETFD, FD_CLOEXEC);
@@ -574,53 +576,37 @@ int setup_tun(struct openconnect_info *vpninfo)
 	return 0;
 }
 
+static struct pkt *out_pkt;
+
 int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 {
-	unsigned char buf[2000];
-	int len;
 	int work_done = 0;
-	int moredata = 1;
-
-	if (FD_ISSET(vpninfo->tun_fd, &vpninfo->select_rfds)) {
-		while (moredata) {
-			unsigned char *pkt = buf;
-#ifdef __sun__
-			/* On Solaris, if we're actually using tuntap and *not* just
-			   passing packets through a UNIX socket to a child process,
-			   we have to use getmsg() to ensure that we only get *one*
-			   packet at a time. */
-			if (!vpninfo->script_tun) {
-				struct strbuf strb;
-				int flags = 0;
-				int ret;
-				
-				strb.buf = buf;
-				strb.maxlen = sizeof(buf);
-				strb.len = 0;
-				ret = getmsg(vpninfo->tun_fd, NULL, &strb, &flags);
-				if (ret < 0)
-					break;
-
-				if (!(ret & MOREDATA))
-					moredata = 0;
-				len = strb.len;
-			} else
-#endif /* __sun__ ... */
-			{
-				len = read(vpninfo->tun_fd, buf, sizeof(buf));
-				if (len <= 0)
-					break;
-			}
+	int prefix_size = 0;
 
 #ifdef TUN_HAS_AF_PREFIX
-			if (!vpninfo->script_tun) {
-				pkt += 4;
-				len -= 4;
-			}
+	if (!vpninfo->script_tun)
+		prefix_size = sizeof(int);
 #endif
-			if (queue_new_packet(&vpninfo->outgoing_queue, pkt,
-					     len))
+
+	if (FD_ISSET(vpninfo->tun_fd, &vpninfo->select_rfds)) {
+		while (1) {
+			int len = vpninfo->mtu;
+
+			if (!out_pkt) {
+				out_pkt = malloc(sizeof(struct pkt) + len);
+				if (!out_pkt) {
+					vpn_progress(vpninfo, PRG_ERR, "Allocation failed\n");
+					break;
+				}
+			}
+
+			len = read(vpninfo->tun_fd, out_pkt->data - prefix_size, len + prefix_size);
+			if (len <= prefix_size)
 				break;
+			out_pkt->len = len - prefix_size;
+
+			queue_packet(&vpninfo->outgoing_queue, out_pkt);
+			out_pkt = NULL;
 
 			work_done = 1;
 			vpninfo->outgoing_qlen++;
@@ -668,10 +654,15 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 #endif
 		vpninfo->incoming_queue = this->next;
 
-		if (write(vpninfo->tun_fd, data, len) < 0 &&
-		    errno == ENOTCONN) {
-			vpninfo->quit_reason = "Client connection terminated";
-			return 1;
+		if (write(vpninfo->tun_fd, data, len) < 0) {
+			/* Handle death of "script" socket */
+			if (vpninfo->script_tun && errno == ENOTCONN) {
+				vpninfo->quit_reason = "Client connection terminated";
+				return 1;
+			}
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to write incoming packet: %s\n"),
+				     strerror(errno));
 		}
 		free(this);
 	}
@@ -684,21 +675,14 @@ void shutdown_tun(struct openconnect_info *vpninfo)
 	if (vpninfo->script_tun) {
 		kill(vpninfo->script_tun, SIGHUP);
 	} else {
-		if (vpninfo->vpnc_script) {
-			setenv("reason", "disconnect", 1);
-			if (system(vpninfo->vpnc_script) == -1) {
-				vpn_progress(vpninfo, PRG_ERR,
-					     _("Failed to spawn script '%s': %s\n"),
-					     vpninfo->vpnc_script,
-					     strerror(errno));
-			}
-		}
+		script_config_tun(vpninfo, "disconnect");
 #ifdef __sun__
-		if (ioctl(vpninfo->ip_fd, I_PUNLINK, vpninfo->tun_muxid) < 0)
-			perror(_("ioctl(I_PUNLINK)"));
-
 		close(vpninfo->ip_fd);
 		vpninfo->ip_fd = -1;
+		if (vpninfo->ip6_fd != -1) {
+			close(vpninfo->ip6_fd);
+			vpninfo->ip6_fd = -1;
+		}
 #endif
 	}
 
