@@ -1,6 +1,6 @@
 #!/bin/sh
 
-v="v3.15"
+v="v3.20"
 
 if [ -d .git ] && tag=`git describe --tags`; then
 	v="$tag"
@@ -9,12 +9,11 @@ if [ -d .git ] && tag=`git describe --tags`; then
 	git update-index --refresh --unmerged > /dev/null
 
 	# Does the index show uncommitted changes?
-	if ! git diff-index --exit-code HEAD > /dev/null; then
+	git diff-index --exit-code HEAD > /dev/null || \
 		v="$v"-dirty
-	fi
 else
 	v="$v"-unknown
 fi
 
-echo "char openconnect_version[] = \"$v\";" > $1
+echo "const char *openconnect_version_str = \"$v\";" > $1
 echo "New version: $v"
